DOMAIN: str = "api.nebius.cloud:443"
DEFAULT_CONFIG_DIR: str = "~/.nebius"
DEFAULT_CONFIG_FILE: str = "config.yaml"
DEFAULT_CREDENTIALS_FILE: str = "credentials.yaml"
TOKEN_ENV = "NEBIUS_IAM_TOKEN"  # noqa: S105
PROFILE_ENV = "NEBIUS_PROFILE"
